/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
/* This file is the UNIX implementation file for finding information about
 * interfaces.  See its header file, of course, and dhcpmain.c for where it's
 * used.
 * A note on its implementation: the number of ifdef's is a bit to look at.
 * There are three methods for determining the MAC address of an interface:
 * - a specific IOCTL (Linux)
 * - looking in the ARP cache (at least Solaris)
 * - using the sysctl interface (FreeBSD and MacOSX)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <sys/ioctl.h>

#if HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif

#if HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif

#if HAVE_NET_IF_DL_H
#include <net/if_dl.h>
#endif

#if HAVE_NET_IF_TYPES_H
#include <net/if_types.h>
#endif

#include "plstr.h"   /* for PL_strdup */
#include "ifdetect.h"

#if HAVE_STRUCT_SOCKADDR_SA_LEN
#define ifreq_len(ifr) \
 PR_MAX(sizeof(struct ifreq), sizeof((ifr)->ifr_name)+(ifr)->ifr_addr.sa_len)
#else
#define ifreq_len(ifr) sizeof(struct ifreq)
#endif

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

#ifndef PR_INADDR_NONE
#define PR_INADDR_NONE PR_UINT32(0xffffffff)
#endif

static PRBool isUpBCastInterface(int sock, const char *name)
{
    PRBool ret;
    struct ifreq ifr;

    NS_ENSURE_TRUE(name, PR_FALSE);

    PL_strncpyz(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(sock, SIOCGIFFLAGS, &ifr)) {
        perror("ioctl SIOCGIFFLAGS");
        ret = 0;
    }
    else
        ret = (ifr.ifr_flags & IFF_UP) && (ifr.ifr_flags & IFF_BROADCAST);
    return ret;
}

static PRUint32 getIPAddr(int fd, const char *name)
{
    PRUint32 ret;
    struct ifreq ifr;

    NS_ENSURE_TRUE(name, PR_INADDR_ANY);

    PL_strncpyz(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFADDR, &ifr)) {
        perror("ioctl SIOCGIFADDR");
        ret = PR_INADDR_ANY;
    }
    else
        memcpy(&ret, ifr.ifr_addr.sa_data + 2, sizeof(PRUint32));
    return ret;
}

static PRUint32 getBCastAddr(int fd, const char *name)
{
    PRUint32 ret;
    struct ifreq ifr;

    NS_ENSURE_TRUE(name, PR_INADDR_NONE);

    PL_strncpyz(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFBRDADDR, &ifr)) {
        perror("ioctl SIOCGIFBRDADDR");
        ret = PR_INADDR_NONE;
    }
    else
        memcpy(&ret, ifr.ifr_addr.sa_data + 2, sizeof(PRUint32));
    return ret;
}

#if defined (SIOCGIFHWADDR)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(struct ifreq));
    PL_strncpyz(ifr.ifr_name, name, IFNAMSIZ);
    if ((ioctl(fd, SIOCGIFHWADDR, &ifr)))
        perror("ioctl SIOCGIFHWADDR");
    else {
        switch (ifr.ifr_hwaddr.sa_family)
        {
            case ARPHRD_ETHER:
            case ARPHRD_FDDI:
            case ARPHRD_IEEE802: // 802.2 Ethernet && Token Ring
            case ARPHRD_IEEE802_TR: // also Token Ring?
                info->hwAddrLen = ETH_ALEN;
                info->hwType = InterfaceHWTypeEthernet;
                break;
            default:
                info->hwAddrLen = PR_MIN(MAX_DHCP_HWADDR,
                 sizeof(ifr.ifr_hwaddr.sa_data));
                info->hwType = InterfaceHWTypeUnknown;
        }
        memcpy(info->hwAddr, ifr.ifr_hwaddr.sa_data, info->hwAddrLen);
    }
}
#elif defined (SIOCGARP)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
    struct arpreq arp;
    struct sockaddr_in *saddr;
    int ret;

    memset(&arp, 0, sizeof(struct arpreq));
    arp.arp_pa.sa_family = AF_INET;
    saddr = (struct sockaddr_in *)&arp; // proto addr is first member of arp
    saddr->sin_family = AF_INET;
    saddr->sin_addr.s_addr = info->ip;
    if ((ret = ioctl(fd, SIOCGARP, &arp)))
        perror("ioctl SIOCGARP");
    else {
        // TODO:  heh:  who said it was ethernet?
        info->hwType = InterfaceHWTypeEthernet;
        info->hwAddrLen = ETH_ALEN;
        memcpy(info->hwAddr, &arp.arp_ha.sa_data[0], ETH_ALEN);
    }
}
#elif defined (HAVE_SYS_SYSCTL_H) && defined (HAVE_NET_IF_DL_H)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
    struct if_msghdr *ifm;
    struct sockaddr_dl *sdl;
    u_char *p, *buf;
    size_t len;
    int mib[] = { CTL_NET, AF_ROUTE, 0, AF_LINK, NET_RT_IFLIST, 0 };

    if (sysctl(mib, 6, nsnull, &len, nsnull, 0) < 0)
        return;

    buf = (u_char *)PR_Malloc(len);
    if (!buf)
        return;

    if (sysctl(mib, 6, buf, &len, nsnull, 0) < 0) {
        PR_Free(buf);
        return;
    }
    for (p = buf; p < buf + len; p += ifm->ifm_msglen) {
        ifm = (struct if_msghdr *)p;
        sdl = (struct sockaddr_dl *)(ifm + 1);

        if (ifm->ifm_type != RTM_IFINFO || (ifm->ifm_addrs & RTA_IFP) == 0)
            continue;

        if (sdl->sdl_family != AF_LINK || sdl->sdl_nlen == 0 ||
         memcmp(sdl->sdl_data, name, PR_MAX(sdl->sdl_nlen, strlen(name))) != 0)
            continue;

        info->hwAddrLen = PR_MIN(MAX_DHCP_HWADDR, sdl->sdl_alen);
#if defined(HAVE_NET_IF_TYPES_H)
        switch (sdl->sdl_type)
        {
            case IFT_ETHER:
            case IFT_FDDI:
            case IFT_ISO88024: // Token Bus
            case IFT_ISO88025: // Token Ring
                info->hwType = InterfaceHWTypeEthernet;
                break;
            default:
                info->hwType = InterfaceHWTypeUnknown;
        }
#else
        // default if we don't know
        info->hwType = InterfaceHWTypeEthernet;
#endif
        memcpy(info->hwAddr, LLADDR(sdl), info->hwAddrLen);
        break;
    }
    PR_Free(buf);
}
#endif

static struct InterfaceInfo *getInterfaceInfo(const char *name, PRUint32 ip)
{
    struct InterfaceInfo *info;
    int fd;

    NS_ENSURE_TRUE(name, nsnull);

    fd = socket (PF_INET, SOCK_DGRAM, 0);
    if (fd != -1) {
        info = (struct InterfaceInfo *)PR_Calloc(sizeof(struct InterfaceInfo),
         1);
        if (info) {
            info->name = PL_strdup(name);
            info->ip = ip;
            info->bcastAddr = getBCastAddr(fd, name);
            getMacAddress(info, fd, name);
        }
        close(fd);
    }
    else {
        perror("socket");
        info = nsnull;
    }
    return info;
}

struct InterfaceInfo *findDHCPInterfaces(void)
{
    const PRUint32 InterfaceIncr = 4;
    int fd = socket (PF_INET, SOCK_DGRAM, 0);
    struct InterfaceInfo *interfaces = nsnull;

    if (fd != -1) {
        int ret, ndx;
        struct ifconf ifc;
        caddr_t ifPtr;

        ret = 0;
        ndx = 0;
        ifc.ifc_buf = nsnull;
        // there is no way to know the interface count beforehand,
        // so we need to loop again and again upping our max each time
        // until returned < max
        do {
            ndx += InterfaceIncr;
            if (ifc.ifc_buf)
                PR_Free(ifc.ifc_buf);
            ifc.ifc_len = sizeof(struct ifreq) * ndx;
            ifc.ifc_buf = (char *)PR_Malloc((PRUint32)ifc.ifc_len);
            if ((ret = ioctl(fd, SIOCGIFCONF, &ifc)) < 0)
                perror("ioctl SIOCGIFCONF");
        } while (ret == 0 && ifc.ifc_len == (sizeof(struct ifreq) * ndx));

        ifPtr = ifc.ifc_buf;
        while (ifPtr < ifc.ifc_buf + ifc.ifc_len) {
            char ifName[IFNAMSIZ];
            PRUint32 ip;

            memcpy(ifName, ifPtr, IFNAMSIZ);
            if (isUpBCastInterface(fd, ifName) && (ip = getIPAddr(fd, ifName))
             != PR_INADDR_ANY) {
                struct InterfaceInfo *newIF = getInterfaceInfo(ifName, ip);

                newIF->next = interfaces;
                interfaces = newIF;
            }
            ifPtr += ifreq_len((struct ifreq *)ifPtr);
        }
        if (ifc.ifc_buf)
            PR_Free(ifc.ifc_buf);
        close(fd);
    }
    else
        perror("socket");
    return interfaces;
}
