/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <stdio.h>
#include <windef.h>
#include <winbase.h>
#include <iphlpapi.h>
#include "nsDebug.h" /* for NS_ASSERTION etc. */
#include "plstr.h"   /* for PL_strdup */
#include "prio.h"    /* for PR_INADDR_ANY */
#include "ifdetect.h"

typedef DWORD (WINAPI *GetAdaptersInfoFunc)(PIP_ADAPTER_INFO, PULONG);
typedef DWORD (WINAPI *GetInterfaceInfoFunc)(PIP_INTERFACE_INFO, PULONG);
typedef DWORD (WINAPI *GetIfTableFunc)(PMIB_IFTABLE, PULONG, BOOL);
typedef DWORD (WINAPI *GetIpAddrTableFunc)(PMIB_IPADDRTABLE, PULONG, BOOL);

static HMODULE gIPHelper = nsnull;
static GetAdaptersInfoFunc gGetAdaptersInfo = nsnull;
static GetInterfaceInfoFunc gGetInterfaceInfo = nsnull;
static GetIfTableFunc gGetIfTable = nsnull;
static GetIpAddrTableFunc gGetIpAddrTable = nsnull;
static OSVERSIONINFO gOSInfo;

static void initLibrary(void)
{
    memset(&gOSInfo, 0, sizeof(gOSInfo));
    gOSInfo.dwOSVersionInfoSize = sizeof(gOSInfo);
    GetVersionEx(&gOSInfo);
    if (!gIPHelper)
        gIPHelper = LoadLibraryA("iphlpapi.dll");
    if (gIPHelper) {
        gGetAdaptersInfo = (GetAdaptersInfoFunc)GetProcAddress(gIPHelper,
         "GetAdaptersInfo");
        gGetInterfaceInfo = (GetInterfaceInfoFunc)GetProcAddress(gIPHelper,
         "GetInterfaceInfo");
        gGetIfTable = (GetIfTableFunc)GetProcAddress(gIPHelper,
         "GetIfTable");
        gGetIpAddrTable = (GetIpAddrTableFunc)GetProcAddress(gIPHelper,
         "GetIpAddrTable");
    }
}

static void freeLibrary(void)
{
    if (gIPHelper) {
        FreeLibrary(gIPHelper);
        gIPHelper = nsnull;
        gGetAdaptersInfo = nsnull;
        gGetInterfaceInfo = nsnull;
        gGetIfTable = nsnull;
        gGetIpAddrTable = nsnull;
    }
}

static char *getAdapterNameByIndex(DWORD index, PIP_INTERFACE_INFO map)
{
    char *ret = nsnull;

    NS_ENSURE_TRUE(map, nsnull);

    for (LONG i = 0; i < map->NumAdapters && !ret; i++) {
        if (map->Adapter[i].Index == index) {
            if (gOSInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
                ret = PL_strdup((char *)map->Adapter[i].Name);
            else {
                char buf[MAX_ADAPTER_NAME];
                WCHAR *p1;
                char *p2;

                for (p1 = map->Adapter[i].Name, p2 = buf;
                 p1 - map->Adapter[i].Name < sizeof(buf) - 1 && *p1;
                 p1++, p2++) {
                    // yes, I'm shortening unicode chars--this is the adapter
                    // name, which isn't meant to be localized
                    *p2 = (char)*p1;
                }
                *p2 = '\0';
                ret = PL_strdup(buf);
            }
        }
    }
    return ret;
}

#define IS_VALID_IF_TYPE(type) \
 ((type) != MIB_IF_TYPE_LOOPBACK && (type) != MIB_IF_TYPE_PPP && \
  (type) != MIB_IF_TYPE_SLIP)

static InterfaceHWType mapIFType(DWORD type)
{
    InterfaceHWType ret;

    switch (type)
    {
        case MIB_IF_TYPE_ETHERNET:
        case MIB_IF_TYPE_TOKENRING:
        case MIB_IF_TYPE_FDDI:
            ret = InterfaceHWTypeEthernet;
            break;
        default:
            ret = InterfaceHWTypeUnknown;
    }
    return ret;
}

static PRBool isUpValidIFRow(PMIB_IFROW row)
{
    if (!row)
        return PR_FALSE;

    return row->dwOperStatus >= MIB_IF_OPER_STATUS_CONNECTED &&
     IS_VALID_IF_TYPE(row->dwType);
}

static PMIB_IPADDRROW findIpEntry(PMIB_IPADDRTABLE table, DWORD index)
{
    PMIB_IPADDRROW ret = nsnull;

    if (!table)
        return nsnull;

    for (DWORD i = 0; i < table->dwNumEntries && !ret; i++)
        if (table->table[i].dwIndex == index)
            ret = &table->table[i];
    return ret;
}

/* Wrapper to GetInterfaceInfo.  Free return value with PR_Free.
 */
static PIP_INTERFACE_INFO getInterfaceInfoWrapper(void)
{
    PIP_INTERFACE_INFO map = nsnull;
    ULONG mapLen = 0;
    DWORD apiReturn;

    if (!gGetInterfaceInfo)
        return nsnull;

    apiReturn = gGetInterfaceInfo(map, &mapLen);
    if (apiReturn != ERROR_INSUFFICIENT_BUFFER || mapLen == 0)
        return nsnull;

    map = (PIP_INTERFACE_INFO)PR_Malloc(mapLen);
    apiReturn = gGetInterfaceInfo(map, &mapLen);
    if (apiReturn != ERROR_SUCCESS) {
        PR_Free(map);
        map = nsnull;
    }
    return map;
}

/* The NT4-specific version, which doesn't use the unsupported GetAdaptersInfo.
 * Still depends on the ip helper API, so doesn't work everywhere.
 */
static struct InterfaceInfo *findDHCPInterfacesNT(void)
{
    struct InterfaceInfo *interfaces = nsnull;
    PMIB_IPADDRTABLE ipTable = nsnull;
    PMIB_IFTABLE ifTable = nsnull;
    PIP_INTERFACE_INFO map = nsnull;
    ULONG ipTableSize = 0, ifTableSize = 0;
    DWORD apiReturn;

    if (!gGetIfTable || !gGetIpAddrTable)
        goto exit;

    apiReturn = gGetIpAddrTable(ipTable, &ipTableSize, FALSE);
    if (apiReturn != ERROR_INSUFFICIENT_BUFFER || ipTableSize == 0)
        goto exit;

    ipTable = (PMIB_IPADDRTABLE)PR_Malloc(ipTableSize);
    apiReturn = gGetIpAddrTable(ipTable, &ipTableSize, FALSE);
    if (apiReturn != NO_ERROR)
        goto exit;

    apiReturn = gGetIfTable(ifTable, &ifTableSize, FALSE);
    if (apiReturn != ERROR_INSUFFICIENT_BUFFER || ifTableSize == 0)
        goto exit;

    ifTable = (PMIB_IFTABLE)PR_Malloc(ifTableSize);
    apiReturn = gGetIfTable(ifTable, &ifTableSize, FALSE);
    if (apiReturn != NO_ERROR)
        goto exit;

    map = getInterfaceInfoWrapper();

    for (DWORD ifNdx = 0; ifNdx < ifTable->dwNumEntries; ifNdx++) {
        PMIB_IFROW ifRow = &ifTable->table[ifNdx];

        if (isUpValidIFRow(ifRow)) {
            PMIB_IPADDRROW ipRow = findIpEntry(ipTable, ifRow->dwIndex);

            if (ipRow && ipRow->dwAddr) {
                struct InterfaceInfo *newIF = (struct InterfaceInfo *)
                 PR_Calloc(1, sizeof(struct InterfaceInfo));

                if (newIF) {
                    DWORD mask = ipRow->dwMask;

                    newIF->name = getAdapterNameByIndex(ifRow->dwIndex, map);
                    newIF->hwType = mapIFType(ifRow->dwType);
                    newIF->hwAddrLen = PR_MIN(MAX_DHCP_HWADDR,
                     ifRow->dwPhysAddrLen);
                    memcpy(newIF->hwAddr, ifRow->bPhysAddr, newIF->hwAddrLen);
                    newIF->ip = ipRow->dwAddr;
                    newIF->bcastAddr = (newIF->ip & mask) | ~mask;
                    newIF->next = interfaces;
                    interfaces = newIF;
                }
            }
        }
    }

exit:
    if (ipTable)
        PR_Free(ipTable);
    if (ifTable)
        PR_Free(ifTable);
    if (map)
        PR_Free(map);
    return interfaces;
}

/* Win98/ME and Win2K+ version, depends on GetAdaptersInfo.  No Win95 support
 * I'm afraid.
 */
static struct InterfaceInfo *findDHCPInterfacesDefault(void)
{
    struct InterfaceInfo *interfaces = nsnull;
    PIP_ADAPTER_INFO adapters = nsnull, ptr = nsnull;
    PIP_INTERFACE_INFO map = nsnull;
    ULONG adaptersLen = 0;
    DWORD apiReturn;

    if (!gGetAdaptersInfo)
        goto exit;

    apiReturn = gGetAdaptersInfo(adapters, &adaptersLen);
    if (apiReturn != ERROR_BUFFER_OVERFLOW || adaptersLen == 0)
        goto exit;

    adapters = (PIP_ADAPTER_INFO)PR_Malloc(adaptersLen);
    apiReturn = gGetAdaptersInfo(adapters, &adaptersLen);
    if (apiReturn != NO_ERROR)
        goto exit;

    if (gOSInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
        map = getInterfaceInfoWrapper();

    for (ptr = adapters; ptr; ptr = ptr->Next) {
        PRNetAddr addr;
        PRStatus rv;

        rv = PR_StringToNetAddr(ptr->IpAddressList.IpAddress.String, &addr);
        if (ptr->DhcpEnabled && IS_VALID_IF_TYPE(ptr->Type) &&
         NS_SUCCEEDED(rv) &&
         addr.inet.family == PR_AF_INET && addr.inet.ip != PR_INADDR_ANY) {
            struct InterfaceInfo *newIF = (struct InterfaceInfo *)
             PR_Calloc(1, sizeof(struct InterfaceInfo));

            if (newIF) {
                PRUint32 mask;

                if (gOSInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
                    newIF->name = getAdapterNameByIndex(ptr->Index, map);
                else
                    newIF->name = PL_strdup(ptr->AdapterName);
                newIF->hwType = mapIFType(ptr->Type);
                newIF->hwAddrLen = PR_MIN(MAX_DHCP_HWADDR, ptr->AddressLength);
                memcpy(newIF->hwAddr, ptr->Address, newIF->hwAddrLen);
                newIF->ip = addr.inet.ip; // was converted above
                rv = PR_StringToNetAddr(ptr->IpAddressList.IpMask.String,
                 &addr);
                if (NS_SUCCEEDED(rv) && addr.inet.family == PR_AF_INET)
                    mask = addr.inet.ip;
                else
                    mask = PR_INADDR_ANY;
                newIF->bcastAddr = (newIF->ip & mask) | ~mask;
                rv = PR_StringToNetAddr(ptr->DhcpServer.IpAddress.String,
                 &addr);
                if (NS_SUCCEEDED(rv) && addr.inet.family == PR_AF_INET)
                    newIF->serverAddr = addr.inet.ip;
                else
                    newIF->serverAddr = PR_INADDR_ANY;
                newIF->next = interfaces;
                interfaces = newIF;
            }
        }
    }

exit:
    if (map)
        PR_Free(map);
    if (adapters)
        PR_Free(adapters);
    return interfaces;
}

struct InterfaceInfo *findDHCPInterfaces(void)
{
    struct InterfaceInfo *interfaces;

    initLibrary();
    if (gOSInfo.dwPlatformId == VER_PLATFORM_WIN32_NT &&
     gOSInfo.dwMajorVersion == 4)
        interfaces = findDHCPInterfacesNT();
    else
        interfaces = findDHCPInterfacesDefault();
    freeLibrary();
    return interfaces;
}
